# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
General Level 1 data readers for ASCAT data in all formats.
Not specific to distributor.
"""
import os
import numpy as np
from datetime import datetime

from pygeobase.io_base import ImageBase
from pygeobase.object_base import Image

import ascat.read_native.eps_native as eps_native
import ascat.read_native.bufr as bufr
import ascat.read_native.nc as nc
import ascat.read_native.hdf5 as h5
from ascat.base import ASCAT_MultiTemporalImageBase
from ascat.math import db2lin, lin2db, hamming_window

byte_nan = np.iinfo(np.byte).min
ubyte_nan = np.iinfo(np.ubyte).max
uint8_nan = np.iinfo(np.uint8).max
uint16_nan = np.iinfo(np.uint16).max
uint32_nan = np.iinfo(np.uint32).max
float32_nan = np.finfo(np.float32).min
float64_nan = np.finfo(np.float64).min
long_nan = np.iinfo(np.int32).min
int_nan = np.iinfo(np.int16).min


class AscatL1Image(ImageBase):
    """
    General Level 1b Image for ASCAT data
    """

    def __init__(self, *args, **kwargs):
        """
        Initialization of i/o object.
        """
        super(AscatL1Image, self).__init__(*args, **kwargs)

    def read(self, timestamp=None, file_format=None, native=False, **kwargs):
        """
        Read ASCAT swath files.

        Parameters
        ----------
        timestamp : str, optional
            Time stamp of file (default: None).
        file_format : str, optional
            File format: '.nat', '.nc', '.bfr', '.h5' (default: None).
            If None file format will be guessed based on the file ending.
        native : bool, optional
            Return native or generic data set format (default: False).
            The main difference is that in the original native format fields
            are called differently.

        Returns
        -------
        img : pygeobase.object_base.Image
            ASCAT swath image.
        """
        if file_format is None:
            file_format = get_file_format(self.filename)

        if file_format in [".nat", ".nat.gz"]:
            img = eps_native.AscatL1bEPSImage(self.filename).read(timestamp)
            if not native:
                img = eps2generic(img)

        elif file_format in [".nc", ".nc.gz"]:
            img = nc.AscatL1NcFile(self.filename).read(timestamp)
            if not native:
                img = nc2generic(img)

        elif file_format in [".bfr", ".bfr.gz", ".buf", "buf.gz"]:
            img = bufr.AscatL1BufrFile(self.filename).read(timestamp)
            if not native:
                img = bfr2generic(img)

        elif file_format in [".h5", ".h5.gz"]:
            img = h5.AscatL1H5File(self.filename).read(timestamp)
            if not native:
                img = hdf2generic(img)

        else:
            raise RuntimeError(
                "Format not found, please indicate the file_format. "
                "[\".nat\", \".nc\", \".bfr\", \".h5\"]")

        return img

    def read_masked_data(self, usable_flag=2, land_flag=0.095, **kwargs):
        """
        Read ASCAT swath files and mask unusable observations.

        Parameters
        ----------
        usable_flag : int, optional
            Allowed usable flag (default: 2).
        land_flag : float, optional
            Sea/land flag threshold (optional: 0.095)

        Returns
        -------
        img : pygeobase.object_base.Image
            ASCAT swath image with masking applied.
        """
        orbit = self.read(**kwargs)

        if type(orbit) is dict:
            img = {'img1': {}, 'img2': {}, 'img3': {},
                   'img4': {}, 'img5': {}, 'img6': {}}

            for single_img in orbit:

                valid = np.ones(
                    orbit[single_img].data[orbit[
                        single_img].data.dtype.names[0]].shape, dtype=np.bool)

                valid = (valid & (orbit[single_img].data[
                    'usable_flag'] < usable_flag))

                valid = (valid & (orbit[single_img].data[
                    'land_flag'] > land_flag))

                valid_num = orbit[single_img].data['jd'][valid].shape[0]
                masked_data = get_template_ASCATL1B_SZF(valid_num)

                for key in orbit[single_img].data.dtype.names:
                    masked_data[key] = orbit[single_img].data[key][valid]

                img[single_img] = Image(orbit[single_img].lon[valid],
                                        orbit[single_img].lat[valid],
                                        masked_data,
                                        orbit[single_img].metadata,
                                        orbit[single_img].timestamp,
                                        timekey='jd')
        else:
            valid = np.ones(
                orbit.data[orbit.data.dtype.names[0]].shape, dtype=np.bool)

            for b in ['f', 'm', 'a']:
                valid = (valid & (orbit.data['usable_flag' + b] < usable_flag))
                valid = (valid & (orbit.data['land_flag' + b] > land_flag))

            valid_num = orbit.data['jd'][valid].shape[0]
            masked_data = get_template_ASCATL1B_SZX(valid_num)

            for key in orbit.data.dtype.names:
                masked_data[key] = orbit.data[key][valid]

            img = Image(orbit.lon[valid], orbit.lat[valid], masked_data,
                        orbit.metadata, orbit.timestamp,
                        timekey='jd')

        return img

    def resample_data(self, data, index, distance, window_radius, **kwargs):
        """
        Weight selected data based on neighborhood (index) and
        distance information.

        Parameters
        ----------
        data : numpy.ndarray
            Selected swath data subset.
        index : numpy.ndarray
            Index information.
        distance : numpy.ndarray
            Distance information.
        window_radius : float
            Size of window.

        Returns
        -------
        res_orbit : numpy.ndarray
            Re-sampled orbit data.
        """
        # target template
        template = get_resample_template_ASCATL1B_SZX()
        res_orbit = np.repeat(template, index.shape[0])

        # get weights
        weights, _ = hamming_window(window_radius, distance)
        # total_weights = np.nansum(weights, axis=1)

        # resample backscatter
        sigmas = ['sigf', 'sigm', 'siga']

        for sigma in sigmas:
            data[sigma][data[sigma] == float32_nan] = np.nan

            weights_exc = weights.copy()
            weights_exc[np.isnan(data[sigma])[index]] = np.nan
            total_weights_exc = np.nansum(weights, axis=1)
            res_orbit[sigma] = lin2db(np.nansum(db2lin(data[sigma])[index]
                                                * weights,
                                                axis=1)
                                      / total_weights_exc)
            res_orbit[sigma][np.isnan(res_orbit[sigma])] = float32_nan
            res_orbit[sigma][np.nansum(weights_exc, axis=1) == 0] = float32_nan

        # resample measurement geometry
        measgeos = ['incf', 'incm', 'inca', 'azif', 'azim', 'azia']

        for mg in measgeos:
            data[mg][data[mg] == float32_nan] = np.nan

            weights_exc = weights.copy()
            weights_exc[np.isnan(data[mg])[index]] = np.nan
            total_weights_exc = np.nansum(weights, axis=1)
            res_orbit[mg] = (np.nansum(data[mg][index] * weights, axis=1)
                             / total_weights_exc)
            res_orbit[mg][np.isnan(res_orbit[mg])] = float32_nan
            res_orbit[mg][np.nansum(weights_exc, axis=1) == 0] = float32_nan

        # noise estimate
        noise = ['kpf', 'kpm', 'kpa']

        for n in noise:
            data[n][data[n] == float32_nan] = np.nan

            weights_exc = weights.copy()
            weights_exc[np.isnan(data[n])[index]] = np.nan
            total_weights_exc = np.nansum(weights, axis=1)
            res_orbit[n] = (np.nansum(data[n][index] * weights, axis=1)
                            / total_weights_exc)
            res_orbit[n][np.isnan(res_orbit[n])] = float32_nan
            res_orbit[n][np.nansum(weights_exc, axis=1) == 0] = float32_nan

        # nearest neighbour resampling values
        nnResample = ['jd', 'sat_id', 'abs_line_nr', 'abs_orbit_nr',
                      'node_num', 'line_num', 'swath', 'as_des_pass']

        # index of min. distance is equal to 0 because of kd-tree usage
        for nn in nnResample:
            res_orbit[nn] = data[nn][index][:, 0]

        # set as_des_pass as string
        # res_orbit['dir'][data['as_des_pass'][index][:, 0] == 1] = 'D'
        # res_orbit['dir'][data['as_des_pass'][index][:, 0] == 0] = 'A'

        # set number of measurements for resampling
        res_orbit['num_obs'] = np.sum(distance != np.inf, axis=1)

        return res_orbit

    def write(self, *args, **kwargs):
        """
        Write function.
        """
        pass

    def flush(self):
        """
        Flush function.
        """
        pass

    def close(self):
        """
        Close function.
        """
        pass


class AscatL1Bufr(ASCAT_MultiTemporalImageBase):
    """
    Class for reading multiple ASCAT level1 images in bufr format.

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        if the files are stored in folders by month then please specify the
        string that should be used in datetime.datetime.strftime
    day_search_str: string, optional
        to provide an iterator over all images of a day the method
        _get_possible_timestamps looks for all available images on a day on the
        harddisk.
        Default: '*-ASCA-*-NA-*-%Y%m%d*.bfr'
    file_search_str: string, optional
        this string is used to find a bufr file by the exact date.
        Default: '*-ASCA-*-NA-*-{datetime}*.bfr'
    datetime_format: string, optional
        datetime format by which {datetime} will be replaced in file_search_str
        Default: %Y%m%d%H%M%S
    msg_name_lookup: dict, optional
        Dictionary mapping bufr msg number to parameter name.
        See bufr.AscatL1BufrFile
    eo_portal : boolean optional
        If your data is from the EUMETSAT EO portal you can set this flag to
        True. This way the the datetime can automatically be read from the
        filename. Otherwise it needs the filename_datetime_format class
        variable set correctly.
    """

    def __init__(self, path, month_path_str='',
                 day_search_str='*-ASCA-*-NA-*-%Y%m%d*-*-*.bfr',
                 file_search_str='*-ASCA-*-NA-*-{datetime}.*.bfr',
                 datetime_format='%Y%m%d%H%M%S',
                 filename_datetime_format=(26, 40, '%Y%m%d%H%M%S'),
                 msg_name_lookup=None, eo_portal=False):

        self.path = path
        self.month_path_str = month_path_str
        self.day_search_str = day_search_str
        self.file_search_str = file_search_str
        self.datetime_format = datetime_format
        self.filename_datetime_format = filename_datetime_format
        self.eo_portal = eo_portal

        super(AscatL1Bufr, self).__init__(
            path, AscatL1Image, subpath_templ=[month_path_str],
            fname_templ=file_search_str, datetime_format=datetime_format,
            exact_templ=False, ioclass_kws={'msg_name_lookup':
                                            msg_name_lookup})

    def _get_orbit_start_date(self, filename):
        """
        Returns the datetime of the file.

        Parameters
        ----------
        filename : full name (including the path) of the file

        Returns
        -------
        dates : datetime object
            datetime from the filename
        """
        # if your data comes from the EUMETSAT EO Portal this function can
        if self.eo_portal is True:
            filename_base = os.path.basename(filename)
            fln_spl = filename_base.split('-')[5]
            fln_datetime = fln_spl.split('.')[0]
            return datetime.strptime(fln_datetime, self.datetime_format)

        else:
            orbit_start_str = os.path.basename(filename)[
                self.filename_datetime_format[0]:
                self.filename_datetime_format[1]]
            return datetime.strptime(orbit_start_str,
                                     self.filename_datetime_format[2])


class AscatL1Eps(ASCAT_MultiTemporalImageBase):
    """
    Class for reading multiple ASCAT level1 images in eps format.

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        if the files are stored in folders by month then please specify the
        string that should be used in datetime.datetime.strftime
    day_search_str: string, optional
        to provide an iterator over all images of a day the method
        _get_possible_timestamps looks for all available images on a day on the
        harddisk.
        Default: 'ASCA_*_*_*_%Y%m%d*_*_*_*_*.nat'
    file_search_str: string, optional
        this string is used to find a bufr file by the exact date.
        Default: 'ASCA_*_*_*_{datetime}Z_*_*_*_*.nat'
    datetime_format: string, optional
        datetime format by which {datetime} will be replaced in file_search_str
        Default: %Y%m%d%H%M%S
    eo_portal : boolean optional
        If your data is from the EUMETSAT EO portal you can set this flag to
        True. This way the the datetime can automatically be read from the
        filename. Otherwise it needs the filename_datetime_format class
        variable set correctly.
    """

    def __init__(self, path, month_path_str='',
                 day_search_str='ASCA_*_*_*_%Y%m%d*_*_*_*_*.nat',
                 file_search_str='ASCA_*_*_*_{datetime}Z_*_*_*_*.nat',
                 datetime_format='%Y%m%d%H%M%S',
                 filename_datetime_format=(16, 30, '%Y%m%d%H%M%S'),
                 eo_portal=False):

        self.path = path
        self.month_path_str = month_path_str
        self.day_search_str = day_search_str
        self.file_search_str = file_search_str
        self.datetime_format = datetime_format
        self.filename_datetime_format = filename_datetime_format
        self.eo_portal = eo_portal

        super(AscatL1Eps, self).__init__(
            path, AscatL1Image, subpath_templ=[month_path_str],
            fname_templ=file_search_str, datetime_format=datetime_format,
            exact_templ=False)


class AscatL1Hdf5(ASCAT_MultiTemporalImageBase):
    """
    Class for reading multiple ASCAT level1 images in hdf5 format.

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        if the files are stored in folders by month then please specify the
        string that should be used in datetime.datetime.strftime
    day_search_str: string, optional
        to provide an iterator over all images of a day the method
        _get_possible_timestamps looks for all available images on a day on the
        harddisk.
        Default: 'ASCA_*_*_*_%Y%m%d*_*_*_*_*.h5'
    file_search_str: string, optional
        this string is used to find a bufr file by the exact date.
        Default: 'ASCA_*_*_*_{datetime}Z_*_*_*_*.h5'
    datetime_format: string, optional
        datetime format by which {datetime} will be replaced in file_search_str
        Default: %Y%m%d%H%M%S
    eo_portal : boolean optional
        If your data is from the EUMETSAT EO portal you can set this flag to
        True. This way the the datetime can automatically be read from the
        filename. Otherwise it needs the filename_datetime_format class
        variable set correctly.
    """

    def __init__(self, path, month_path_str='',
                 day_search_str='ASCA_*_*_*_%Y%m%d*_*_*_*_*.h5',
                 file_search_str='ASCA_*_*_*_{datetime}Z_*_*_*_*.h5',
                 datetime_format='%Y%m%d%H%M%S',
                 filename_datetime_format=(16, 30, '%Y%m%d%H%M%S'),
                 eo_portal=False):

        self.path = path
        self.month_path_str = month_path_str
        self.day_search_str = day_search_str
        self.file_search_str = file_search_str
        self.datetime_format = datetime_format
        self.filename_datetime_format = filename_datetime_format
        self.eo_portal = eo_portal

        super(AscatL1Hdf5, self).__init__(
            path, AscatL1Image, subpath_templ=[month_path_str],
            fname_templ=file_search_str, datetime_format=datetime_format,
            exact_templ=False)


class AscatL1Nc(ASCAT_MultiTemporalImageBase):
    """
    Class for reading multiple ASCAT level1 images in nc format.

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        if the files are stored in folders by month then please specify the
        string that should be used in datetime.datetime.strftime
    day_search_str: string, optional
        to provide an iterator over all images of a day the method
        _get_possible_timestamps looks for all available images on a day on the
        harddisk.
        Default: 'W_XX-*_EUMP_%Y%m%d*.nc'
    file_search_str: string, optional
        this string is used to find a bufr file by the exact date.
        Default: 'W_XX-*_EUMP_{datetime}*.nc'
    datetime_format: string, optional
        datetime format by which {datetime} will be replaced in file_search_str
        Default: %Y%m%d%H%M%S
    msg_name_lookup: dict, optional
        Dictionary mapping nc msg number to parameter name.
        See nc.AscatL1NcFile
    eo_portal : boolean optional
        If your data is from the EUMETSAT EO portal you can set this flag to
        True. This way the the datetime can automatically be read from the
        filename. Otherwise it needs the filename_datetime_format class
        variable set correctly.
    """

    def __init__(self, path, month_path_str='',
                 day_search_str='W_XX-*_EUMP_%Y%m%d*.nc',
                 file_search_str='W_XX-*_EUMP_{datetime}*.nc',
                 datetime_format='%Y%m%d%H%M%S',
                 filename_datetime_format=(62, 76, '%Y%m%d%H%M%S'),
                 msg_name_lookup=None, eo_portal=False):

        self.path = path
        self.month_path_str = month_path_str
        self.day_search_str = day_search_str
        self.file_search_str = file_search_str
        self.datetime_format = datetime_format
        self.filename_datetime_format = filename_datetime_format
        self.eo_portal = eo_portal

        super(AscatL1Nc, self).__init__(
            path, AscatL1Image, subpath_templ=[month_path_str],
            fname_templ=file_search_str, datetime_format=datetime_format,
            exact_templ=False, ioclass_kws={'msg_name_lookup':
                                            msg_name_lookup})


def get_file_format(filename):
    """
    Try to guess the file format from the extension.

    Parameters
    ----------
    filename : str
        File name.

    Returns
    -------
    file_format : str
        File format indicator.
    """
    if os.path.splitext(filename)[1] == '.gz':
        file_format = os.path.splitext(os.path.splitext(filename)[0])[1]
    else:
        file_format = os.path.splitext(filename)[1]
    return file_format


def eps2generic(native_img):
    """
    Convert the native eps Image into a generic one.
    Use different convert-function for szf and szx data

    Parameters
    ----------
    native_img : pygeobase.object_base.Image
        Native image.

    Returns
    -------
    img : pygeobase.object_base.Image
        Converted image.
    """
    if type(native_img) is dict:
        img = eps_dict2generic(native_img)
    else:
        img = eps_img2generic(native_img)

    return img


def eps_img2generic(native_img):
    """
    Convert single eps image into a generic one.

    Parameters
    ----------
    native_img : pygeobase.object_base.Image
        Native image.

    Returns
    -------
    img : pygeobase.object_base.Image
        Generic image.
    """
    n_records = native_img.lat.shape[0]
    generic_data = get_template_ASCATL1B_SZX(n_records)

    fields = [('jd', 'jd', None),
              ('sat_id', None, None),
              ('abs_line_nr', None, None),
              ('abs_orbit_nr', None, None),
              ('node_num', 'NODE_NUM', None),
              ('line_num', 'LINE_NUM', None),
              ('as_des_pass', 'AS_DES_PASS', None),
              ('swath', 'SWATH INDICATOR', byte_nan),
              ('azif', 'f_AZI_ANGLE_TRIP', int_nan),
              ('azim', 'm_AZI_ANGLE_TRIP', int_nan),
              ('azia', 'a_AZI_ANGLE_TRIP', int_nan),
              ('incf', 'f_INC_ANGLE_TRIP', uint16_nan),
              ('incm', 'm_INC_ANGLE_TRIP', uint16_nan),
              ('inca', 'a_INC_ANGLE_TRIP', uint16_nan),
              ('sigf', 'f_SIGMA0_TRIP', long_nan),
              ('sigm', 'm_SIGMA0_TRIP', long_nan),
              ('siga', 'a_SIGMA0_TRIP', long_nan),
              ('kpf', 'f_KP', uint16_nan),
              ('kpm', 'm_KP', uint16_nan),
              ('kpa', 'a_KP', uint16_nan),
              ('kpf_quality', 'f_F_KP', byte_nan),
              ('kpm_quality', 'm_F_KP', byte_nan),
              ('kpa_quality', 'a_F_KP', byte_nan),
              ('land_flagf', 'f_F_LAND', uint16_nan),
              ('land_flagm', 'm_F_LAND', uint16_nan),
              ('land_flaga', 'a_F_LAND', uint16_nan),
              ('usable_flagf', 'f_F_USABLE', byte_nan),
              ('usable_flagm', 'm_F_USABLE', byte_nan),
              ('usable_flaga', 'a_F_USABLE', byte_nan)]

    for field in fields:
        if field[1] is None:
            continue

        if field[2] is not None:
            valid_mask = (native_img.data[field[1]] != field[2])
            generic_data[field[0]][valid_mask] = native_img.data[field[1]][
                valid_mask]
        else:
            generic_data[field[0]] = native_img.data[field[1]]

    if 'ABS_LINE_NUMBER' in native_img.data:
        generic_data['abs_line_nr'] = native_img.data['ABS_LINE_NUMBER']

    fields = [('sat_id', 'SPACECRAFT_ID'),
              ('abs_orbit_nr', 'ORBIT_START')]

    for field in fields:
        generic_data[field[0]] = native_img.metadata[field[1]].repeat(
            n_records)

    # convert sat_id (spacecraft id) to the intern definition
    sat_id_lut = np.array([0, 4, 3, 5])
    generic_data['sat_id'] = sat_id_lut[generic_data['sat_id']]

    img = Image(native_img.lon, native_img.lat, generic_data,
                native_img.metadata, native_img.timestamp,
                timekey='jd')

    return img


def eps_dict2generic(native_img):
    """
    Convert dict of eps Images into a dict with generic Images

    Parameters
    ----------
    native_img : pygeobase.object_base.Image
        Native image.

    Returns
    -------
    img : list of pygeobase.object_base.Image
        Generic images.
    """
    img = {'img1': {}, 'img2': {}, 'img3': {},
           'img4': {}, 'img5': {}, 'img6': {}}

    for szf_img in native_img:
        n_records = native_img[szf_img].lat.shape[0]
        generic_data = get_template_ASCATL1B_SZF(n_records)

        fields = [('jd', 'jd'),
                  ('sat_id', None),
                  ('beam_number', 'BEAM_NUMBER'),
                  ('abs_orbit_nr', None),
                  ('as_des_pass', 'AS_DES_PASS'),
                  ('azi', 'AZI_ANGLE_FULL'),
                  ('inc', 'INC_ANGLE_FULL'),
                  ('sig', 'SIGMA0_FULL'),
                  ('land_frac', 'LAND_FRAC'),
                  ('flagfield_rf1', 'FLAGFIELD_RF1'),
                  ('flagfield_rf2', 'FLAGFIELD_RF2'),
                  ('flagfield_pl', 'FLAGFIELD_PL'),
                  ('flagfield_gen1', 'FLAGFIELD_GEN1'),
                  ('flagfield_gen2', 'FLAGFIELD_GEN2'),
                  ('land_flag', 'F_LAND'),
                  ('usable_flag', 'F_USABLE')]

        for field in fields:
            if field[1] is None:
                continue
            generic_data[field[0]] = native_img[szf_img].data[field[1]]

        fields = [('sat_id', 'SPACECRAFT_ID'),
                  ('abs_orbit_nr', 'ORBIT_START')]
        for field in fields:
            generic_data[field[0]] = native_img[szf_img].metadata[
                field[1]].repeat(n_records)

        # convert sat_id (spacecraft id) to the intern definition
        sat_id_lut = np.array([0, 4, 3, 5])
        generic_data['sat_id'] = sat_id_lut[generic_data['sat_id']]

        img[szf_img] = Image(native_img[szf_img].lon,
                             native_img[szf_img].lat,
                             generic_data,
                             native_img[szf_img].metadata,
                             native_img[szf_img].timestamp,
                             timekey='jd')
    return img


def nc2generic(native_img):
    """
    Convert the native nc Image into a generic one.

    Parameters
    ----------
    native_img : pygeobase.object_base.Image
        Native image.

    Returns
    -------
    img : pygeobase.object_base.Image
        Generic images.
    """
    n_records = native_img.lat.shape[0]
    generic_data = get_template_ASCATL1B_SZX(n_records)

    fields = [('jd', 'jd'),
              ('sat_id', None),
              ('abs_line_nr', None),
              ('abs_orbit_nr', None),
              ('node_num', 'node_num'),
              ('line_num', 'line_num'),
              ('as_des_pass', 'as_des_pass'),
              ('swath', 'swath_indicator'),
              ('azif', 'f_azi_angle_trip'),
              ('azim', 'm_azi_angle_trip'),
              ('azia', 'a_azi_angle_trip'),
              ('incf', 'f_inc_angle_trip'),
              ('incm', 'm_inc_angle_trip'),
              ('inca', 'a_inc_angle_trip'),
              ('sigf', 'f_sigma0_trip'),
              ('sigm', 'm_sigma0_trip'),
              ('siga', 'a_sigma0_trip'),
              ('kpf', 'f_kp'),
              ('kpm', 'm_kp'),
              ('kpa', 'a_kp'),
              ('kpf_quality', 'f_f_kp'),
              ('kpm_quality', 'm_f_kp'),
              ('kpa_quality', 'a_f_kp'),
              ('land_flagf', 'f_f_land'),
              ('land_flagm', 'm_f_land'),
              ('land_flaga', 'a_f_land'),
              ('usable_flagf', 'f_f_usable'),
              ('usable_flagm', 'm_f_usable'),
              ('usable_flaga', 'a_f_usable')]

    for field in fields:
        if field[1] is None:
            continue

        if type(native_img.data[field[1]]) == np.ma.core.MaskedArray:
            valid_mask = ~native_img.data[field[1]].mask
            generic_data[field[0]][valid_mask] = native_img.data[field[1]][
                valid_mask]
        else:
            generic_data[field[0]] = native_img.data[field[1]]

    if 'abs_line_number' in native_img.data:
        generic_data['abs_line_nr'] = native_img.data['abs_line_number']

    fields = [('sat_id', 'sat_id'), ('abs_orbit_nr', 'orbit_start')]

    for field in fields:
        generic_data[field[0]] = np.repeat(native_img.metadata[field[1]],
                                           n_records)

    # convert sat_id (spacecraft id) to the intern definition
    sat_id_lut = np.array([0, 4, 3, 5])
    generic_data['sat_id'] = sat_id_lut[generic_data['sat_id']]

    img = Image(native_img.lon, native_img.lat, generic_data,
                native_img.metadata, native_img.timestamp,
                timekey='jd')

    return img


def bfr2generic(native_img):
    """
    Convert the native bfr Image into a generic one.

    Parameters
    ----------
    native_img : pygeobase.object_base.Image
        Native image.

    Returns
    -------
    img : pygeobase.object_base.Image
        Generic images.
    """
    n_records = native_img.lat.shape[0]
    generic_data = get_template_ASCATL1B_SZX(n_records)

    fields = [('jd', 'jd', None),
              ('sat_id', 'Satellite Identifier', None),
              ('abs_line_nr', None, None),
              ('abs_orbit_nr', 'Orbit Number', None),
              ('node_num', 'Cross-Track Cell Number', None),
              ('line_num', 'line_num', None),
              ('as_des_pass', 'as_des_pass', None),
              ('swath', 'swath_indicator', None),
              ('azif', 'f_Antenna Beam Azimuth', 1.7e+38),
              ('azim', 'm_Antenna Beam Azimuth', 1.7e+38),
              ('azia', 'a_Antenna Beam Azimuth', 1.7e+38),
              ('incf', 'f_Radar Incidence Angle', 1.7e+38),
              ('incm', 'm_Radar Incidence Angle', 1.7e+38),
              ('inca', 'a_Radar Incidence Angle', 1.7e+38),
              ('sigf', 'f_Backscatter', 1.7e+38),
              ('sigm', 'm_Backscatter', 1.7e+38),
              ('siga', 'a_Backscatter', 1.7e+38),
              ('kpf', 'f_Radiometric Resolution (Noise Value)', 1.7e+38),
              ('kpm', 'm_Radiometric Resolution (Noise Value)', 1.7e+38),
              ('kpa', 'a_Radiometric Resolution (Noise Value)', 1.7e+38),
              ('kpf_quality', 'f_ASCAT KP Estimate Quality', None),
              ('kpm_quality', 'm_ASCAT KP Estimate Quality', None),
              ('kpa_quality', 'a_ASCAT KP Estimate Quality', None),
              ('land_flagf', 'f_ASCAT Land Fraction', None),
              ('land_flagm', 'm_ASCAT Land Fraction', None),
              ('land_flaga', 'a_ASCAT Land Fraction', None),
              ('usable_flagf', 'f_ASCAT Sigma-0 Usability', None),
              ('usable_flagm', 'm_ASCAT Sigma-0 Usability', None),
              ('usable_flaga', 'a_ASCAT Sigma-0 Usability', None)]

    kp_vars = ['kpf', 'kpm', 'kpa']

    for field in fields:
        if field[1] is None:
            continue

        if field[2] is not None:
            valid_mask = (native_img.data[field[1]] != field[2])
            if field[0] in kp_vars:
                generic_data[field[0]][valid_mask] = \
                    native_img.data[field[1]][valid_mask] / 100
            else:
                generic_data[field[0]][valid_mask] = \
                    native_img.data[field[1]][valid_mask]
        else:
            generic_data[field[0]] = native_img.data[field[1]]

    # convert sat_id (spacecraft id) to the intern definition
    sat_id_lut = np.array([0, 0, 0, 4, 3, 5])
    generic_data['sat_id'] = sat_id_lut[generic_data['sat_id']]

    img = Image(native_img.lon, native_img.lat, generic_data,
                native_img.metadata, native_img.timestamp,
                timekey='jd')

    return img


def hdf2generic(native_img):
    """
    Convert the native nc Image into a generic one.

    Parameters
    ----------
    native_img : list of pygeobase.object_base.Image
        Native image.

    Returns
    -------
    img : list of pygeobase.object_base.Image
        Generic images.
    """

    img = {'img1': {}, 'img2': {}, 'img3': {},
           'img4': {}, 'img5': {}, 'img6': {}}

    for szf_img in native_img:
        n_records = native_img[szf_img].lat.shape[0]
        generic_data = get_template_ASCATL1B_SZF(n_records)

        fields = [('jd', 'jd'),
                  ('sat_id', None),
                  ('beam_number', 'BEAM_NUMBER'),
                  ('abs_orbit_nr', None),
                  ('as_des_pass', 'AS_DES_PASS'),
                  ('azi', 'AZI_ANGLE_FULL'),
                  ('inc', 'INC_ANGLE_FULL'),
                  ('sig', 'SIGMA0_FULL'),
                  ('land_frac', 'LAND_FRAC'),
                  ('flagfield_rf1', 'FLAGFIELD_RF1'),
                  ('flagfield_rf2', 'FLAGFIELD_RF2'),
                  ('flagfield_pl', 'FLAGFIELD_PL'),
                  ('flagfield_gen1', 'FLAGFIELD_GEN1'),
                  ('flagfield_gen2', 'FLAGFIELD_GEN2'),
                  ('land_flag', 'F_LAND'),
                  ('usable_flag', 'F_USABLE')
                  ]
        for field in fields:
            if field[1] is None:
                continue
            generic_data[field[0]] = native_img[szf_img].data[field[1]]

        fields = [('sat_id', 'SPACECRAFT_ID'),
                  ('abs_orbit_nr', 'ORBIT_START')]
        for field in fields:
            generic_data[field[0]] = np.repeat(native_img[szf_img].metadata[
                field[1]], n_records)

        # convert sat_id (spacecraft id) to the intern definition
        sat_id_lut = np.array([0, 4, 3, 5])
        generic_data['sat_id'] = sat_id_lut[generic_data['sat_id']]

        img[szf_img] = Image(native_img[szf_img].lon,
                             native_img[szf_img].lat,
                             generic_data,
                             native_img[szf_img].metadata,
                             native_img[szf_img].timestamp,
                             timekey='jd')

    return img


def get_template_ASCATL1B_SZX(n=1):
    """
    Generic Level 1b SZX template.

    Parameters
    ----------
    n : int, optional
        Number of records (default: 1).

    Returns
    -------
    records : numpy.ndarray
        Array filled with default values.
    """
    metadata = {'temp_name': 'ASCATL1B_SZX'}

    struct = np.dtype([('jd', np.float64),
                       ('sat_id', np.byte),
                       ('abs_line_nr', np.uint32),
                       ('abs_orbit_nr', np.uint32),
                       ('node_num', np.uint8),
                       ('line_num', np.uint16),
                       ('as_des_pass', np.byte),
                       ('swath', np.byte),
                       ('azif', np.float32),
                       ('azim', np.float32),
                       ('azia', np.float32),
                       ('incf', np.float32),
                       ('incm', np.float32),
                       ('inca', np.float32),
                       ('sigf', np.float32),
                       ('sigm', np.float32),
                       ('siga', np.float32),
                       ('kpf', np.float32),
                       ('kpm', np.float32),
                       ('kpa', np.float32),
                       ('kpf_quality', np.float32),
                       ('kpm_quality', np.float32),
                       ('kpa_quality', np.float32),
                       ('land_flagf', np.uint8),
                       ('land_flagm', np.uint8),
                       ('land_flaga', np.uint8),
                       ('usable_flagf', np.uint8),
                       ('usable_flagm', np.uint8),
                       ('usable_flaga', np.uint8)], metadata=metadata)

    record = np.array([(float64_nan, byte_nan, uint32_nan, uint32_nan,
                        uint8_nan, uint16_nan, byte_nan, byte_nan, float32_nan,
                        float32_nan, float32_nan, float32_nan, float32_nan,
                        float32_nan, float32_nan, float32_nan, float32_nan,
                        float32_nan, float32_nan, float32_nan, float32_nan,
                        float32_nan, float32_nan, uint8_nan, uint8_nan,
                        uint8_nan, uint8_nan, uint8_nan, uint8_nan)],
                      dtype=struct)

    return np.repeat(record, n)


def get_template_ASCATL1B_SZF(n=1):
    """
    Generic Level 1b SZF template.

    Parameters
    ----------
    n : int, optional
        Number of records (default: 1).

    Returns
    -------
    records : numpy.ndarray
        Array filled with default values.
    """
    metadata = {'temp_name': 'ASCATL1B_SZF'}

    struct = np.dtype([('jd', np.float64),
                       ('sat_id', np.byte),
                       ('beam_number', np.byte),
                       ('abs_orbit_nr', np.uint32),
                       ('as_des_pass', np.byte),
                       ('azi', np.float32),
                       ('inc', np.float32),
                       ('sig', np.float32),
                       ('land_frac', np.float32),
                       ('flagfield_rf1', np.uint8),
                       ('flagfield_rf2', np.uint8),
                       ('flagfield_pl', np.uint8),
                       ('flagfield_gen1', np.uint8),
                       ('flagfield_gen2', np.uint8),
                       ('land_flag', np.uint8),
                       ('usable_flag', np.uint8)], metadata=metadata)

    record = np.array([(float64_nan, byte_nan, byte_nan, uint32_nan, byte_nan,
                        float32_nan, float32_nan, float32_nan, float32_nan,
                        uint8_nan, uint8_nan, uint8_nan, uint8_nan, uint8_nan,
                        uint8_nan, uint8_nan)], dtype=struct)

    return np.repeat(record, n)


def get_resample_template_ASCATL1B_SZX(n=1):
    """
    Generic Level 1b SZX template.

    Parameters
    ----------
    n : int, optional
        Number of records (default: 1).

    Returns
    -------
    records : numpy.ndarray
        Array filled with default values.
    """
    metadata = {'temp_name': 'ASCRS010'}

    struct = np.dtype([('jd', np.float64),
                       ('sat_id', np.byte),
                       ('abs_line_nr', np.uint32),
                       ('abs_orbit_nr', np.uint32),
                       ('node_num', np.uint8),
                       ('line_num', np.uint16),
                       ('as_des_pass', np.byte),
                       ('swath', np.byte),
                       ('azif', np.float32),
                       ('azim', np.float32),
                       ('azia', np.float32),
                       ('incf', np.float32),
                       ('incm', np.float32),
                       ('inca', np.float32),
                       ('sigf', np.float32),
                       ('sigm', np.float32),
                       ('siga', np.float32),
                       ('kpf', np.float32),
                       ('kpm', np.float32),
                       ('kpa', np.float32),
                       ('num_obs', np.uint16)], metadata=metadata)

    record = np.array([(float64_nan, byte_nan, uint32_nan, uint32_nan,
                        uint8_nan, uint16_nan, byte_nan, byte_nan, float32_nan,
                        float32_nan, float32_nan, float32_nan, float32_nan,
                        float32_nan, float32_nan, float32_nan, float32_nan,
                        float32_nan, float32_nan, float32_nan, uint16_nan)],
                      dtype=struct)

    return np.repeat(record, n)



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Tests for level 1 reader.
"""
import numpy as np
import numpy.testing as nptest
import unittest
import os
from datetime import datetime

import ascat.level1 as level1

float32_nan = np.finfo(np.float32).min


class Test_AscatL1Image(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(
            os.path.dirname(__file__), 'ascat_test_data', 'eumetsat',
            'ASCAT_generic_reader_data')

        name_b = os.path.join(data_path, 'bufr',
                              'M02-ASCA-ASCSZR1B0200-NA-9.1-20100609013900.000000000Z-20130824233100-1280350.bfr')
        name_e = os.path.join(data_path, 'eps_nat',
                              'ASCA_SZR_1B_M02_20100609013900Z_20100609032058Z_R_O_20130824233100Z.nat')
        name_n = os.path.join(data_path, 'nc',
                              'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOPA+ASCAT_C_EUMP_20100609013900_18872_eps_o_125_l1.nc')

        name_e11 = os.path.join(data_path, 'eps_nat',
                                'ASCA_SZR_1B_M02_20071212071500Z_20071212085659Z_R_O_20081225063118Z.nat')
        name_e_szf = os.path.join(data_path, 'eps_nat',
                                  'ASCA_SZF_1B_M01_20180611041800Z_20180611055959Z_N_O_20180611050637Z.nat')
        name_h = os.path.join(data_path, 'hdf5',
                              'ASCA_SZF_1B_M01_20180611041800Z_20180611055959Z_N_O_20180611050637Z.h5')

        self.image_bufr = level1.AscatL1Image(name_b)
        self.image_eps = level1.AscatL1Image(name_e)
        self.image_nc = level1.AscatL1Image(name_n)

        self.image_eps_fmv11 = level1.AscatL1Image(name_e11)

        self.image_e_szf = level1.AscatL1Image(name_e_szf)
        self.image_h5_szf = level1.AscatL1Image(name_h)

    def tearDown(self):
        self.image_nc = None
        self.image_bufr = None
        self.image_eps = None
        self.image_eps_fmv11 = None
        self.image_eps_szf = None
        self.image_h5_szf = None

    def test_image_reading_szx_all_formats(self):
        self.reader_bufr = self.image_bufr.read()
        self.reader_eps = self.image_eps.read()
        self.reader_nc = self.image_nc.read()

        nptest.assert_allclose(self.reader_bufr.lat, self.reader_eps.lat,
                               atol=1e-4)
        nptest.assert_allclose(self.reader_eps.lat, self.reader_nc.lat,
                               atol=1e-4)
        nptest.assert_allclose(self.reader_nc.lat, self.reader_bufr.lat,
                               atol=1e-4)

        nptest.assert_allclose(self.reader_bufr.lon, self.reader_eps.lon,
                               atol=1e-4)
        nptest.assert_allclose(self.reader_eps.lon, self.reader_nc.lon,
                               atol=1e-4)
        nptest.assert_allclose(self.reader_nc.lon, self.reader_bufr.lon,
                               atol=1e-4)

        matching = ['jd', 'sat_id', 'abs_line_nr', 'abs_orbit_nr', 'node_num',
                    'line_num', 'as_des_pass', 'swath', 'azif', 'azim', 'azia',
                    'incf', 'incm', 'inca', 'sigf', 'sigm', 'siga', 'kpf',
                    'kpm', 'kpa', 'kpf_quality', 'kpm_quality', 'kpa_quality',
                    'land_flagf', 'land_flagm', 'land_flaga', 'usable_flagf',
                    'usable_flagm', 'usable_flaga']

        # lists with no data fields or different definition
        bufr_none = ['abs_line_nr', 'abs_orbit_nr', 'as_des_pass']
        sig = ['sigf', 'sigm', 'siga']

        # BUFR files contain less accurate data so we only compare to one 0.1
        # accuracy.
        for field in matching:
            if field not in bufr_none:
                # BUFR filters sigma values below -50 dB so for comparison we
                # have to mask those values in nc and eps
                if field in sig:
                    sig_mask = (self.reader_eps.data[field] < -50)
                    self.reader_eps.data[field][sig_mask] = float32_nan
                    self.reader_nc.data[field][sig_mask] = float32_nan
                    self.reader_bufr.data[field][sig_mask] = float32_nan

                    nan_mask = (self.reader_nc.data[field] == float32_nan)
                    self.reader_eps.data[field][nan_mask] = float32_nan
                    self.reader_bufr.data[field][nan_mask] = float32_nan

                nptest.assert_allclose(self.reader_bufr.data[field],
                                       self.reader_eps.data[field], atol=0.1)
                nptest.assert_allclose(self.reader_nc.data[field],
                                       self.reader_bufr.data[field], atol=0.1)

            nptest.assert_allclose(self.reader_eps.data[field],
                                   self.reader_nc.data[field], atol=0.1)

    def test_image_reading_szf_all_formats(self):
        self.reader_eps_szf = self.image_e_szf.read()
        self.reader_hdf5_szf = self.image_h5_szf.read()
        for szf_img in self.reader_eps_szf:
            nptest.assert_allclose(self.reader_eps_szf[szf_img].lat,
                                   self.reader_hdf5_szf[szf_img].lat,
                                   atol=1e-4)

            nptest.assert_allclose(self.reader_eps_szf[szf_img].lon,
                                   self.reader_hdf5_szf[szf_img].lon,
                                   atol=1e-4)

            matching = ['jd', 'sat_id', 'beam_number', 'abs_orbit_nr',
                        'as_des_pass', 'azi', 'inc', 'sig', 'land_frac',
                        'flagfield_rf1', 'flagfield_rf2', 'flagfield_pl',
                        'flagfield_gen1', 'flagfield_gen2',
                        'land_flag', 'usable_flag']

            for field in matching:
                nptest.assert_allclose(
                    self.reader_eps_szf[szf_img].data[field],
                    self.reader_hdf5_szf[szf_img].data[field], atol=0.1)

    def test_image_reading_szx_eps(self):
        self.reader = self.image_eps.read(file_format='.nat')

        lat_should = np.array(
            [68.91681, 69.005196, 69.09337, 69.18132, 69.26905, 69.35655,
             69.443825, 69.53087, 69.61768, 69.704254, 69.79059, 69.87668,
             69.962524, 70.04812, 70.13346, 70.218544, 70.303375, 70.38794,
             70.47224, 70.55626, 70.640015, 70.723495, 70.806694, 70.88961,
             70.97224])

        lon_should = np.array(
            [168.80144, 168.60977, 168.41656, 168.22179, 168.02544, 167.82748,
             167.62794, 167.42676, 167.22394, 167.01947, 166.81332, 166.60548,
             166.39592, 166.18465, 165.97163, 165.75685, 165.5403, 165.32195,
             165.10178, 164.87979, 164.65594, 164.43024, 164.20264, 163.97314,
             163.74171])

        sig_should = np.array(
            [-13.510671, -13.421737, -13.872492, -14.351357, -14.395881,
             -14.382635, -14.860762, -16.108913, -17.354418, -18.86383,
             -18.793966, -18.631758, -18.46626, -18.71435, -19.150038,
             -19.315845, -19.79865, -19.845669, -19.892258, -20.138796,
             -20.151554, -20.154343, -20.165552, -20.013523, -19.238102])

        kp_should = np.array(
            [0.0307, 0.032, 0.051, 0.0696, 0.0703, 0.0584, 0.045, 0.0464,
             0.0615, 0.0477, 0.0323, 0.04, 0.0346, 0.0369, 0.0378, 0.0397,
             0.0341, 0.0399, 0.0418, 0.0408, 0.0421, 0.0347, 0.0424, 0.0451,
             0.0523])

        jd_should = np.array(
            [2455356.56875, 2455356.56875, 2455356.56875,
             2455356.56875, 2455356.56875, 2455356.56875,
             2455356.56875, 2455356.56877169, 2455356.56877169,
             2455356.56877169])

        nptest.assert_allclose(self.reader.lat[:25], lat_should, atol=1e-5)
        nptest.assert_allclose(self.reader.lon[:25], lon_should, atol=1e-5)
        nptest.assert_allclose(self.reader.data['sigf'][:25],
                               sig_should, atol=1e-5)
        nptest.assert_allclose(self.reader.data['kpf'][:25],
                               kp_should, atol=1e-5)
        nptest.assert_allclose(self.reader.data['jd'][75:85],
                               jd_should, atol=1e-5)

    def test_image_reading_szx_eps_fmv11(self):
        self.reader = self.image_eps_fmv11.read()

        lat_should = np.array(
            [61.849445, 61.916786, 61.983864, 62.050674, 62.11722, 62.183495,
             62.2495, 62.315228, 62.380684, 62.44586, 62.51076, 62.57538,
             62.63971, 62.70376, 62.76752, 62.830994, 62.894173, 62.95706,
             63.019653, 63.081947, 63.143944, 63.205635, 63.267025, 63.32811,
             63.388885])

        lon_should = np.array(
            [69.18133, 68.991295, 68.80043, 68.60872, 68.41617, 68.22277,
             68.028534, 67.833435, 67.63749, 67.44069, 67.243034, 67.04452,
             66.84513, 66.64489, 66.44378, 66.2418, 66.03894, 65.83522,
             65.630615, 65.42514, 65.21878, 65.01154, 64.80342, 64.59441,
             64.38452])

        sig_should = np.array(
            [-15.008125, -14.547356, -15.067405, -15.340037, -15.381483,
             -15.085848, -14.620477, -14.200545, -13.873865, -13.29581,
             -12.962119, -12.909232, -12.990307, -13.076723, -13.039384,
             -13.010556, -13.238036, -13.045113, -12.981088, -13.003889,
             -14.009461, -14.633162, -14.706434, -14.042056, -13.7074])

        kp_should = np.array(
            [0.052, 0.0417, 0.0462, 0.0264, 0.0308, 0.0296, 0.0363, 0.0348,
             0.0377, 0.036, 0.0329, 0.0258, 0.0296, 0.0245, 0.0275, 0.0309,
             0.035, 0.0325, 0.0288, 0.0292, 0.0431, 0.0363, 0.0435, 0.0282,
             0.0309])

        jd_should = np.array(
            [2454446.80208333, 2454446.80208333, 2454446.80208333,
             2454446.80208333, 2454446.80208333, 2454446.80208333,
             2454446.80208333, 2454446.80210505, 2454446.80210505,
             2454446.80210505])

        nptest.assert_allclose(self.reader.lat[:25], lat_should, atol=1e-5)
        nptest.assert_allclose(self.reader.lon[:25], lon_should, atol=1e-5)
        nptest.assert_allclose(self.reader.data['sigf'][:25],
                               sig_should, atol=1e-5)
        nptest.assert_allclose(self.reader.data['kpf'][:25],
                               kp_should, atol=1e-5)
        nptest.assert_allclose(self.reader.data['jd'][75:85],
                               jd_should, atol=1e-5)

    def test_image_reading_szf_eps(self):
        self.reader = self.image_e_szf.read()

        lat_should = np.array(
            [64.45502, 64.42318, 64.39127, 64.35929, 64.32724, 64.29512,
             64.262924, 64.23065, 64.19831, 64.16589, 64.1334, 64.10083,
             64.06819, 64.03547, 64.00268, 63.969807, 63.936855, 63.903828,
             63.870724, 63.837536, 63.804276, 63.77093, 63.737507, 63.704002,
             63.67042])

        lon_should = np.array(
            [103.29956, 103.32185, 103.34413, 103.36641, 103.38869,
             103.41095, 103.43322, 103.45548, 103.47774, 103.49999,
             103.52224, 103.54449, 103.566734, 103.588974, 103.61121,
             103.63345, 103.655685, 103.67792, 103.70014, 103.722374,
             103.7446, 103.76682, 103.78904, 103.811264, 103.83348])

        sig_should = np.array(
            [-9.713457, -8.768949, -9.294478, -7.449275, -8.939872,
             -7.893198, -8.570546, -8.934691, -7.851117, -7.782818,
             -8.33993, -7.539894, -7.833797, -8.465893, -8.244121,
             -7.59996, -8.976448, -9.36595, -10.800382, -8.289896,
             -9.127579, -9.410345, -7.238986, -8.335969, -7.897769])

        jd_should = np.array(
            [2458280.67917396, 2458280.67917396, 2458280.67918378,
             2458280.67918378, 2458280.67918378, 2458280.67918378,
             2458280.67918378, 2458280.67918378, 2458280.67918378,
             2458280.67918378])

        nptest.assert_allclose(self.reader['img1'].lat[:25],
                               lat_should, atol=1e-5)
        nptest.assert_allclose(self.reader['img1'].lon[:25],
                               lon_should, atol=1e-5)
        nptest.assert_allclose(self.reader['img1'].data['sig'][:25],
                               sig_should, atol=1e-5)
        nptest.assert_allclose(self.reader['img1'].data['jd'][190:200],
                               jd_should, atol=1e-5)


class Test_AscatL1Bufr(unittest.TestCase):

    def setUp(self):
        self.data_path = os.path.join(
            os.path.dirname(__file__), 'ascat_test_data', 'eumetsat',
            'ASCAT_generic_reader_data', 'bufr')

        self.image_bufr = level1.AscatL1Bufr(self.data_path, eo_portal=True)

    def tearDown(self):
        self.image_bufr = None

    def test_image_reading(self):
        data, meta, timestamp, lon, lat, time_var = self.image_bufr.read(
            datetime(2010, 6, 9, 1, 39))

        assert lon.shape == (267648,)
        assert lat.shape == (267648,)

    def test_get_orbit_start_date(self):
        filename = os.path.join(self.data_path,
                                'M02-ASCA-ASCSZR1B0200-NA-9.1-20100609013900.000000000Z-20130824233100-1280350.bfr')
        orbit_start = self.image_bufr._get_orbit_start_date(filename)
        orbit_start_should = datetime(2010, 6, 9, 1, 39)

        assert orbit_start == orbit_start_should

    def test_tstamp_for_daterange(self):
        tstamps = self.image_bufr.tstamps_for_daterange(datetime(2010, 6, 8),
                                                        datetime(2010, 6, 10))
        tstamps_should = [datetime(2010, 6, 9, 1, 39)]

        assert tstamps == tstamps_should


class Test_AscatL1Eps(unittest.TestCase):

    def setUp(self):
        self.data_path = os.path.join(
            os.path.dirname(__file__), 'ascat_test_data', 'eumetsat',
            'ASCAT_generic_reader_data', 'eps_nat')

        self.image_eps = level1.AscatL1Eps(self.data_path, eo_portal=True)

    def tearDown(self):
        self.image_eps = None

    def test_image_reading(self):
        data, meta, timestamp, lon, lat, time_var = self.image_eps.read(
            datetime(2010, 6, 9, 1, 39))

        assert lon.shape == (267648,)
        assert lat.shape == (267648,)

    def test_image_reading_szf(self):
        szf_dict = self.image_eps.read(
            datetime(2018, 6, 11, 4, 18))

        szf_dict['img1'].lon
        assert szf_dict['img1'].lon.shape == (1383552,)
        assert szf_dict['img1'].lat.shape == (1383552,)

    def test_get_orbit_start_date(self):
        filename = os.path.join(self.data_path,
                                'ASCA_SZR_1B_M02_20100609013900Z_20100609032058Z_R_O_20130824233100Z.nat')
        orbit_start = self.image_eps._get_orbit_start_date(filename)
        orbit_start_should = datetime(2010, 6, 9, 1, 39)

        assert orbit_start == orbit_start_should

    def test_tstamp_for_daterange(self):
        tstamps = self.image_eps.tstamps_for_daterange(
            datetime(2010, 6, 8), datetime(2010, 6, 10))
        tstamps_should = [datetime(2010, 6, 9, 1, 39)]

        assert tstamps == tstamps_should


class Test_AscatL1Nc(unittest.TestCase):

    def setUp(self):
        self.data_path = os.path.join(
            os.path.dirname(__file__), 'ascat_test_data', 'eumetsat',
            'ASCAT_generic_reader_data', 'nc')

        self.image_nc = level1.AscatL1Nc(self.data_path, eo_portal=True)

    def tearDown(self):
        self.image_nc = None

    def test_image_reading(self):
        data, meta, timestamp, lon, lat, time_var = self.image_nc.read(
            datetime(2010, 6, 9, 1, 39))

        assert lon.shape == (267648,)
        assert lat.shape == (267648,)

    def test_get_orbit_start_date(self):
        filename = os.path.join(self.data_path,
                                'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOPA+ASCAT_C_EUMP_20100609013900_18872_eps_o_125_l1.nc')
        orbit_start = self.image_nc._get_orbit_start_date(filename)
        orbit_start_should = datetime(2010, 6, 9, 1, 39)

        assert orbit_start == orbit_start_should

    def test_tstamp_for_daterange(self):
        tstamps = self.image_nc.tstamps_for_daterange(datetime(2010, 6, 8),
                                                      datetime(2010, 6, 10))
        tstamps_should = [datetime(2010, 6, 9, 1, 39)]

        assert tstamps == tstamps_should


class Test_AscatL1Hdf5(unittest.TestCase):

    def setUp(self):
        self.data_path = os.path.join(
            os.path.dirname(__file__), 'ascat_test_data', 'eumetsat',
            'ASCAT_generic_reader_data', 'hdf5')

        self.image_h5 = level1.AscatL1Hdf5(self.data_path, eo_portal=True)

    def tearDown(self):
        self.image_h5 = None

    def test_image_reading(self):
        szf_dict = self.image_h5.read(
            datetime(2018, 6, 11, 4, 18))

        szf_dict['img1'].lon
        assert szf_dict['img1'].lon.shape == (1383552,)
        assert szf_dict['img1'].lat.shape == (1383552,)

    def test_get_orbit_start_date(self):
        filename = os.path.join(self.data_path,
                                'ASCA_SZF_1B_M01_20180611041800Z_20180611055959Z_N_O_20180611050637Z.h5')
        orbit_start = self.image_h5._get_orbit_start_date(filename)
        orbit_start_should = datetime(2018, 6, 11, 4, 18)

        assert orbit_start == orbit_start_should

    def test_tstamp_for_daterange(self):
        tstamps = self.image_h5.tstamps_for_daterange(datetime(2018, 6, 10),
                                                      datetime(2018, 6, 12))
        tstamps_should = [datetime(2018, 6, 11, 4, 18)]

        assert tstamps == tstamps_should



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Tests for level 2 readers.
"""

from datetime import datetime
import numpy as np
import numpy.testing as nptest
import os
import pytest
import unittest
import sys

from ascat.read_native.bufr import AscatL2SsmBufr
from ascat.read_native.bufr import AscatL2SsmBufrChunked
from ascat.read_native.bufr import AscatL2SsmBufrFile
from ascat.read_native.nc import AscatL2SsmNcFile
import ascat.level2 as level2

float32_nan = np.finfo(np.float32).min


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_AscatL2SsmBufr_ioclass_kws(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(os.path.dirname(__file__), 'ascat_test_data',
                                 'hsaf', 'h07')
        self.reader = AscatL2SsmBufr(data_path,
                                     msg_name_lookup={65: 'ssm',
                                                      74: 'ssm mean'})

    def tearDown(self):
        self.reader = None

    def test_offset_getting(self):
        """
        test getting the image offsets for a known day
        2010-05-01
        """
        timestamps = self.reader.tstamps_for_daterange(
            datetime(2010, 5, 1), datetime(2010, 5, 1, 12))
        timestamps_should = [datetime(2010, 5, 1, 8, 33, 1)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read(
            datetime(2010, 5, 1, 8, 33, 1))

        ssm_should = np.array([51.2, 65.6, 46.2, 56.9, 61.4, 61.5, 58.1, 47.1,
                               72.7, 13.8, 60.9, 52.1, 78.5, 57.8, 56.2, 79.8,
                               67.7, 53.8, 86.5, 29.4, 50.6, 88.8, 56.9, 68.9,
                               52.4, 64.4, 81.5, 50.5, 84., 79.6, 47.4, 79.5,
                               46.9, 60.7, 81.3, 52.9, 84.5, 25.5, 79.2, 93.3,
                               52.6, 93.9, 74.4, 91.4, 76.2, 92.5, 80., 88.3,
                               79.1, 97.2, 56.8])

        lats_should = np.array([70.21162, 69.32506, 69.77325, 68.98149,
                                69.12295, 65.20364, 67.89625, 67.79844,
                                67.69112, 67.57446, 67.44865, 67.23221,
                                66.97207, 66.7103, 66.34695, 65.90996,
                                62.72462, 61.95761, 61.52935, 61.09884,
                                60.54359, 65.60223, 65.33588, 65.03098,
                                64.58972, 61.46131, 60.62553, 59.52057,
                                64.27395, 63.80293, 60.6569, 59.72684,
                                58.74838, 63.42774])

        ssm_mean_should = np.array([0.342, 0.397, 0.402, 0.365, 0.349,
                                    0.354, 0.37, 0.36, 0.445, 0.211,
                                    0.394, 0.354, 0.501, 0.361, 0.366,
                                    0.45, 0.545, 0.329, 0.506, 0.229,
                                    0.404, 0.591, 0.348, 0.433, 0.29,
                                    0.508, 0.643, 0.343, 0.519, 0.61,
                                    0.414, 0.594, 0.399, 0.512, 0.681,
                                    0.457, 0.622, 0.396, 0.572, 0.7,
                                    0.302, 0.722, 0.493, 0.747, 0.521,
                                    0.72, 0.578, 0.718, 0.536, 0.704,
                                    0.466]) * 100

        nptest.assert_allclose(lats[25:-1:30], lats_should, atol=1e-5)
        nptest.assert_allclose(data['ssm'][
                               15:-1:20], ssm_should, atol=0.01)
        nptest.assert_allclose(data['ssm mean'][15:-1:20],
                               ssm_mean_should,
                               atol=0.01)


class Test_AscatL2SsmBufrFile(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(
            os.path.dirname(__file__), 'ascat_test_data', 'eumetsat',
            'ASCAT_L2_SM_125', 'bufr', 'Metop_B')
        fname = os.path.join(
            data_path,
            'M01-ASCA-ASCSMR02-NA-5.0-20170220050900.000000000Z-20170220055833-1207110.bfr')
        self.reader = AscatL2SsmBufrFile(fname)

    def tearDown(self):
        self.reader = None

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read()

        ssm_should = np.array(
            [29.2, 30.2, 35.7, 38.6, 37.5, 37.6, 40.5, 44.5, 40.7,
             39.7, 41.5, 38.8, 34.5, 36.8, 39.4, 41.2, 42.4, 42.9,
             39.3, 30.5, 26.7, 26.5, 26.7, 23.9, 26.2])

        lats_should = np.array(
            [64.74398, 64.81854, 64.89284, 64.96688, 65.04066, 65.11416,
             65.18739, 65.26036, 65.33304, 65.40545, 65.47758, 65.54942,
             65.62099, 65.69226, 65.76324, 65.83393, 65.90432, 65.97442,
             66.04422, 66.11371, 66.1829, 66.25177, 66.32034, 66.38859,
             66.45653])

        ssm_mean_should = np.array(
            [36.7, 35.4, 33.4, 32.5, 32.5, 32., 31.2, 29.4, 28.7,
             27.6, 25.8, 25.4, 25.5, 25.3, 24.4, 23.4, 22.3, 21.3,
             20.4, 20.4, 19.9, 19.7, 20.3, 21.5, 22.9])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               :25], ssm_should, atol=1e-5)
        nptest.assert_allclose(data['Mean Surface Soil Moisture'][:25],
                               ssm_mean_should,
                               atol=1e-5)

    def test_image_reading_masked(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read(
            ssm_masked=True)

        ssm_should = np.array(
            [15.6, 10.8, 15.3, 15.9, 19.8, 27., 27.8, 26.8, 28.6,
             35.6, 36., 32.3, 27.6, 31.2, 36.8, 13.4, 18.7, 23.1,
             24.5, 22.1, 17.1, 17.9, 17.8, 21.1, 23.])

        lats_should = np.array(
            [54.27036, 54.3167, 54.36279, 54.40862, 54.45419, 54.49951,
             54.54456, 54.58936, 54.6339, 54.67818, 54.72219, 54.76594,
             54.80943, 54.85265, 54.89561, 56.95692, 56.98178, 57.00631,
             57.03053, 57.05442, 57.07799, 57.10123, 57.12415, 57.14675,
             57.16902])

        ssm_mean_should = np.array(
            [24.4, 22.2, 21., 19.6, 18.7, 19.1, 19.1, 19.9, 19.9,
             20.1, 20.9, 21., 19.6, 17.3, 16.9, 25.6, 25.6, 24.9,
             23.6, 22.9, 22.4, 23.2, 24.1, 24.5, 26.1])

        nptest.assert_allclose(lats[10000:10025], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)']
                               [10000:10025], ssm_should,
                               atol=1e-5)
        nptest.assert_allclose(data['Mean Surface Soil Moisture']
                               [10000:10025], ssm_mean_should,
                               atol=1e-5)


class Test_AscatL2SsmNcFile(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(
            os.path.dirname(__file__), 'ascat_test_data', 'eumetsat',
            'ASCAT_L2_SM_125', 'nc', 'Metop_A')
        fname = os.path.join(
            data_path,
            'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOPA+ASCAT_C_EUMP_20170220041500_53652_eps_o_125_ssm_l2.nc')
        self.reader = AscatL2SsmNcFile(fname)

    def tearDown(self):
        self.reader = None

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read()

        ssm_should = np.array([2.96000004, 0., 0., 0., 0., 0., 0., 0., 0.,
                               1.82999992, 3.32999992, 4.78999996, 4.31999969,
                               2.53999996, 0., 3.83999991, 5.76999998, 1.5,
                               2.44000006, 4.11999989, 2.25999999, 2.65999985,
                               5.5999999, 5.53999996, 4.85999966])

        lats_should = np.array([62.60224, 62.67133, 62.74015, 62.80871, 62.877,
                                62.94502, 63.01276, 63.08024, 63.14743,
                                63.21435, 63.28098, 63.34734, 63.41341,
                                63.47919, 63.54468, 63.60988, 63.67479,
                                63.7394, 63.80372, 63.86773, 63.93144,
                                63.99485, 64.05795, 64.12075, 64.18323])

        ssm_mean_should = np.array([21.26000023, 21.27999878, 21.38999939,
                                    22.43000031, 23.36999893, 24.51000023,
                                    26.01000023, 27.04999924, 26.94999886,
                                    26.63999939, 27.09999847, 27.56999969,
                                    27.43000031, 26.64999962, 26.53999901,
                                    27.48999977, 28.20999908, 28.38999939,
                                    28.79999924, 29.21999931, 30.01000023,
                                    30.97999954, 31.27999878, 31.8599987,
                                    32.05999756])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['soil_moisture'][
                               :25], ssm_should, atol=1e-5)
        nptest.assert_allclose(data['mean_soil_moisture'][:25],
                               ssm_mean_should,
                               atol=1e-5)

    def test_image_reading_masked(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read(
            ssm_masked=True)

        ssm_should = np.array(
            [33.39999771118164, 27.06999969482422,
             20.649999618530273, 18.28999900817871,
             24.229999542236328, 24.939998626708984,
             23.639999389648438, 20.3799991607666,
             14.15999984741211, 10.059999465942383,
             9.539999961853027, 9.019999504089355,
             9.420000076293945, 12.279999732971191,
             21.529998779296875, 33.880001068115234,
             39.57999801635742, 35.34000015258789,
             38.88999938964844, 44.459999084472656,
             46.66999816894531, 40.12999725341797,
             38.39999771118164, 43.959999084472656,
             33.43000030517578])

        lats_should = np.array(
            [65.11197384, 65.17437784, 65.23645384, 65.29819884, 65.35961083,
             65.42068783, 65.48142683, 65.54182483, 65.60187983, 65.66158983,
             65.72095083, 65.77996183, 65.83861883, 68.62952883, 68.66132883,
             68.69261383, 68.72337983, 68.75362483, 68.78334683, 68.81254383,
             68.84121383, 68.86935383, 68.89696283, 68.92403783, 68.95057683])

        ssm_mean_should = np.array([26.85999870300293, 25.90999984741211,
                                    25.670000076293945, 25.81999969482422,
                                    24.65999984741211, 22.6299991607666,
                                    20.389999389648438, 18.94999885559082,
                                    17.68000030517578, 16.28999900817871,
                                    15.130000114440918, 14.739999771118164,
                                    15.5,
                                    26.51999855041504, 31.529998779296875,
                                    36.09000015258789, 40.36000061035156,
                                    42.61000061035156, 45.529998779296875,
                                    47.939998626708984, 47.45000076293945,
                                    44.689998626708984, 41.12999725341797,
                                    37.59000015258789, 33.09000015258789])

        nptest.assert_allclose(lats[50000:50025], lats_should, atol=1e-5)
        nptest.assert_allclose(data['soil_moisture'][50000:50025], ssm_should,
                               atol=1e-5)
        nptest.assert_allclose(data['mean_soil_moisture'][50000:50025],
                               ssm_mean_should,
                               atol=1e-5)


class Test_AscatL2SsmNcFile_vsAscatL2SsmBufrFile(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(
            os.path.dirname(__file__), 'ascat_test_data', 'eumetsat',
            'ASCAT_L2_SM_125')
        fname_nc = os.path.join(
            data_path, 'nc', 'Metop_A',
            'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOPA+ASCAT_C_EUMP_20170220041500_53652_eps_o_125_ssm_l2.nc')
        self.reader_nc = AscatL2SsmNcFile(fname_nc)

        fname_bufr = os.path.join(
            data_path, 'bufr', 'Metop_A',
            'M02-ASCA-ASCSMR02-NA-5.0-20170220041500.000000000Z-20170220055656-1207110.bfr')
        self.reader_bufr = AscatL2SsmBufrFile(fname_bufr)

    def tearDown(self):
        self.reader_nc = None
        self.reader_bufr = None

    def test_image_reading(self):
        data_nc, meta, timestamp, lons_nc, lats_nc, time_var_nc = self.reader_nc.read()
        data_bufr, meta, timestamp, lons_bufr, lats_bufr, time_var_bufr = self.reader_bufr.read()

        nptest.assert_allclose(lats_nc, lats_bufr, atol=1e-4)

        nc_bufr_matching = {
            'slope40': 'Slope At 40 Deg Incidence Angle',
            'sigma40_error': 'Estimated Error In Sigma0 At 40 Deg Incidence Angle',
            'utc_line_nodes': None,
            'jd': 'jd',
            'wet_backscatter': 'Wet Backscatter',
            'swath_indicator': None,
            'frozen_soil_probability': 'Frozen Land Surface Fraction',
            'wetland_flag': 'Inundation And Wetland Fraction',
            # The processing flag definition between BUFR and netCDF is slightly different
            # 'proc_flag1':                'Soil Moisture Processing Flag',
            'proc_flag2': None,
            'abs_line_number': None,
            'sat_track_azi': None,
            'sigma40': 'Backscatter',
            'soil_moisture': 'Surface Soil Moisture (Ms)',
            'soil_moisture_error': 'Estimated Error In Surface Soil Moisture',
            'rainfall_flag': 'Rain Fall Detection',
            'soil_moisture_sensitivity': 'Soil Moisture Sensitivity',
            'corr_flags': 'Soil Moisture Correction Flag',
            'dry_backscatter': 'Dry Backscatter',
            'aggregated_quality_flag': None,
            'mean_soil_moisture': 'Mean Surface Soil Moisture',
            'as_des_pass': None,
            'slope40_error': 'Estimated Error In Slope At 40 Deg Incidence Angle',
            'topography_flag': 'Topographic Complexity',
            'snow_cover_probability': 'Snow Cover'}

        # 'Direction Of Motion Of Moving Observing Platform']
        # BUFR files contain less accurate data so we only compare to one 0.1
        # accuracy.
        for nc_name in nc_bufr_matching:
            bufr_name = nc_bufr_matching[nc_name]
            if bufr_name is None:
                continue

            # flags and probabilities do not have the same NaN value so we mask
            # the invalid values for comparison
            if nc_name in ['snow_cover_probability',
                           'rainfall_flag',
                           'topography_flag',
                           'frozen_soil_probability',
                           'wetland_flag',
                           'snow_cover_probability']:
                valid = np.where(data_nc[nc_name] != 255)
                data_nc[nc_name] = data_nc[nc_name][valid]
                data_bufr[bufr_name] = data_bufr[bufr_name][valid]

            nptest.assert_allclose(data_nc[nc_name],
                                   data_bufr[bufr_name], atol=0.1)


def test_AscatL2SsmBufrChunked():
    data_path = os.path.join(
        os.path.dirname(
            __file__), 'ascat_test_data', 'eumetsat', 'ASCAT_L2_SM_125',
        'PDU', 'Metop_B')
    day_search_str = 'W_XX-EUMETSAT-Darmstadt,SOUNDING+SATELLITE,METOPB+ASCAT_C_EUMP_%Y%m%d*_125_ssm_l2.bin'
    file_search_str = 'W_XX-EUMETSAT-Darmstadt,SOUNDING+SATELLITE,METOPB+ASCAT_C_EUMP_{datetime}*_125_ssm_l2.bin'
    datetime_format = '%Y%m%d%H%M%S'
    filename_datetime_format = (63, 77, '%Y%m%d%H%M%S')
    reader = AscatL2SsmBufrChunked(data_path, month_path_str='',
                                   day_search_str=day_search_str,
                                   file_search_str=file_search_str,
                                   datetime_format=datetime_format,
                                   filename_datetime_format=filename_datetime_format)

    intervals = reader.tstamps_for_daterange(datetime(2017, 2, 20, 5),
                                             datetime(2017, 2, 20, 6))
    data = reader.read(intervals[0])
    assert len(data.metadata.keys()) == 3
    assert data.data['jd'].shape == (23616,)
    assert data.lon.shape == (23616,)
    assert data.lat.shape == (23616,)


class Test_AscatL2Image(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(
            os.path.dirname(__file__), 'ascat_test_data', 'eumetsat',
            'ASCAT_generic_reader_data')
        name_b = os.path.join(data_path, 'bufr',
                              'M01-ASCA-ASCSMO02-NA-5.0-20180612035700.000000000Z-20180612044530-1281300.bfr')
        name_e = os.path.join(data_path, 'eps_nat',
                              'ASCA_SMO_02_M01_20180612035700Z_20180612053856Z_N_O_20180612044530Z.nat')
        name_n = os.path.join(data_path, 'nc',
                              'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOPB+ASCAT_C_EUMP_20180612035700_29742_eps_o_250_ssm_l2.nc')
        self.image_bufr = level2.AscatL2Image(name_b)
        self.image_eps = level2.AscatL2Image(name_e)
        self.image_nc = level2.AscatL2Image(name_n)

    def tearDown(self):
        self.image_nc = None
        self.image_bufr = None
        self.image_eps = None

    def test_image_reading_all_formats(self):
        self.reader_bufr = self.image_bufr.read()
        self.reader_eps = self.image_eps.read()
        self.reader_nc = self.image_nc.read()

        nptest.assert_allclose(self.reader_bufr.lat, self.reader_eps.lat,
                               atol=1e-4)
        nptest.assert_allclose(self.reader_eps.lat, self.reader_nc.lat,
                               atol=1e-4)
        nptest.assert_allclose(self.reader_nc.lat, self.reader_bufr.lat,
                               atol=1e-4)

        nptest.assert_allclose(self.reader_bufr.lon, self.reader_eps.lon,
                               atol=1e-4)
        nptest.assert_allclose(self.reader_eps.lon, self.reader_nc.lon,
                               atol=1e-4)
        nptest.assert_allclose(self.reader_nc.lon, self.reader_bufr.lon,
                               atol=1e-4)

        matching = ['jd', 'sat_id', 'abs_line_nr', 'abs_orbit_nr', 'node_num',
                    'line_num', 'as_des_pass', 'swath', 'azif', 'azim', 'azia',
                    'incf', 'incm', 'inca', 'sigf', 'sigm', 'siga', 'sm',
                    'sm_noise', 'sm_sensitivity', 'sig40', 'sig40_noise',
                    'slope40', 'slope40_noise', 'dry_backscatter',
                    'wet_backscatter', 'mean_surf_sm', 'correction_flag',
                    'processing_flag', 'aggregated_quality_flag',
                    'snow_cover_probability', 'frozen_soil_probability',
                    'innudation_or_wetland', 'topographical_complexity']

        # lists with no data fields
        bufr_none = ['abs_line_nr', 'abs_orbit_nr', 'aggregated_quality_flag']
        nc_none = ['azif', 'azim', 'azia', 'incf', 'incm', 'inca',
                   'sigf', 'sigm', 'siga', 'processing_flag']

        # BUFR files contain less accurate data so we only compare to one 0.1
        # accuracy.
        for field in matching:

            # difference between the files should not be the case
            if field == 'sig40':
                mask = self.reader_nc.data[field] == float32_nan
                self.reader_bufr.data[field][mask] = float32_nan
                self.reader_eps.data[field][mask] = float32_nan

            # difference between the files should not be the case
            if field in ['snow_cover_probability', 'frozen_soil_probability',
                         'innudation_or_wetland', 'topographical_complexity']:
                mask = self.reader_eps.data[field] == float32_nan
                self.reader_nc.data[field][mask] = float32_nan
                self.reader_eps.data[field][mask] = float32_nan

            if field not in bufr_none:
                nptest.assert_allclose(self.reader_bufr.data[field],
                                       self.reader_eps.data[field], atol=0.1)

            if field not in nc_none:
                nptest.assert_allclose(self.reader_eps.data[field],
                                       self.reader_nc.data[field], atol=0.1)

            if field not in bufr_none and field not in nc_none:
                nptest.assert_allclose(self.reader_nc.data[field],
                                       self.reader_bufr.data[field], atol=0.1)

    def test_image_reading_eps(self):
        self.reader = self.image_eps.read()

        sm_should = np.array(
            [69.11, 74.23, 74.12, 75.95, 76.23, 80.74, 83.45, 84.94, 84.28,
             86.33, 86.19, 86.31, 87.64, 87.92, 90.65, 90.52, 89.71, 89.33,
             91.41, 91.89, 94.51, 70.43, 67.75, 60.54, 69.43])

        lat_should = np.array(
            [64.06651, 64.21156, 64.355545, 64.49845, 64.64026, 64.78095,
             64.9205, 65.05891, 65.19613, 65.33216, 65.46697, 65.600555,
             65.73289, 65.86394, 65.9937, 66.12214, 66.249245, 66.374985,
             66.499344, 66.62231, 66.743835, 69.63313, 69.698105, 69.760895,
             69.821495])

        lon_should = np.array(
            [121.95572, 121.564156, 121.16849, 120.76867, 120.36467,
             119.95644, 119.54396, 119.12719, 118.70608, 118.2806,
             117.85073, 117.41643, 116.97765, 116.53439, 116.08661,
             115.63427, 115.17735, 114.715836, 114.24969, 113.77889,
             113.30343, 96.66666, 96.049965, 95.42956, 94.80551])

        mean_surf_sm_should = np.array(
            [77.97, 77.57, 79.2, 78.38, 77.85, 79.81, 80.72, 81.23, 82.43,
             82.11, 81.93, 82.55, 83.41, 81.84, 81.43, 81.28, 80.37, 79.6,
             79.43, 78.02, 77.49, 42.42, 41.69, 42.99, 47.51])

        jd_should = np.array(
            [2458281.66458332, 2458281.66458332, 2458281.66458332,
             2458281.66458332, 2458281.66458332, 2458281.66458332,
             2458281.66458332, 2458281.66462674, 2458281.66462674,
             2458281.66462674])

        nptest.assert_allclose(self.reader.lat[:25], lat_should, atol=1e-5)
        nptest.assert_allclose(self.reader.lon[:25], lon_should, atol=1e-5)
        nptest.assert_allclose(self.reader.data['sm'][:25],
                               sm_should, atol=1e-5)
        nptest.assert_allclose(self.reader.data['mean_surf_sm'][:25],
                               mean_surf_sm_should, atol=1e-5)
        nptest.assert_allclose(self.reader.data['jd'][35:45],
                               jd_should, atol=1e-5)


class Test_AscatL2Bufr(unittest.TestCase):

    def setUp(self):
        self.data_path = os.path.join(
            os.path.dirname(__file__), 'ascat_test_data', 'eumetsat',
            'ASCAT_generic_reader_data', 'bufr')

        self.image_bufr = level2.AscatL2Bufr(self.data_path, eo_portal=True)

    def tearDown(self):
        self.image_bufr = None

    def test_image_reading(self):
        data, meta, timestamp, lon, lat, time_var = self.image_bufr.read(
            datetime(2018, 6, 12, 3, 57))

        assert lon.shape == (68544,)
        assert lat.shape == (68544,)

    def test_get_orbit_start_date(self):
        filename = os.path.join(self.data_path,
                                'M01-ASCA-ASCSMO02-NA-5.0-20180612035700.000000000Z-20180612044530-1281300.bfr')
        orbit_start = self.image_bufr._get_orbit_start_date(filename)
        orbit_start_should = datetime(2018, 6, 12, 3, 57)

        assert orbit_start == orbit_start_should

    def test_tstamp_for_daterange(self):
        tstamps = self.image_bufr.tstamps_for_daterange(datetime(2018, 6, 12),
                                                        datetime(2018, 6, 13))
        tstamps_should = [datetime(2018, 6, 12, 3, 57)]

        assert tstamps == tstamps_should


class Test_AscatL2Eps(unittest.TestCase):

    def setUp(self):
        self.data_path = os.path.join(
            os.path.dirname(__file__), 'ascat_test_data', 'eumetsat',
            'ASCAT_generic_reader_data', 'eps_nat')

        self.image_eps = level2.AscatL2Eps(self.data_path, eo_portal=True)

    def tearDown(self):
        self.image_eps = None

    def test_image_reading(self):
        data, meta, timestamp, lon, lat, time_var = self.image_eps.read(
            datetime(2018, 6, 12, 3, 57))

        assert lon.shape == (68544,)
        assert lat.shape == (68544,)

    def test_get_orbit_start_date(self):
        filename = os.path.join(self.data_path,
                                'ASCA_SMO_02_M01_20180612035700Z_20180612053856Z_N_O_20180612044530Z.nat')
        orbit_start = self.image_eps._get_orbit_start_date(filename)
        orbit_start_should = datetime(2018, 6, 12, 3, 57)

        assert orbit_start == orbit_start_should

    def test_tstamp_for_daterange(self):
        tstamps = self.image_eps.tstamps_for_daterange(datetime(2018, 6, 12),
                                                       datetime(2018, 6, 13))
        tstamps_should = [datetime(2018, 6, 12, 3, 57)]

        assert tstamps == tstamps_should


class Test_AscatL2Nc(unittest.TestCase):

    def setUp(self):
        self.data_path = os.path.join(
            os.path.dirname(__file__), 'ascat_test_data', 'eumetsat',
            'ASCAT_generic_reader_data', 'nc')

        self.image_nc = level2.AscatL2Nc(self.data_path, eo_portal=True)

    def tearDown(self):
        self.image_nc = None

    def test_image_reading(self):
        data, meta, timestamp, lon, lat, time_var = self.image_nc.read(
            datetime(2018, 6, 12, 3, 57))

        assert lon.shape == (68544,)
        assert lat.shape == (68544,)

    def test_get_orbit_start_date(self):
        filename = os.path.join(self.data_path,
                                'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOPB+ASCAT_C_EUMP_20180612035700_29742_eps_o_250_ssm_l2.nc')
        orbit_start = self.image_nc._get_orbit_start_date(filename)
        orbit_start_should = datetime(2018, 6, 12, 3, 57)

        assert orbit_start == orbit_start_should

    def test_tstamp_for_daterange(self):
        tstamps = self.image_nc.tstamps_for_daterange(datetime(2018, 6, 12),
                                                      datetime(2018, 6, 13))
        tstamps_should = [datetime(2018, 6, 12, 3, 57)]

        assert tstamps == tstamps_should



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import warnings
import zipfile
from datetime import datetime

import numpy as np
import pandas as pd

import pygeogrids.grids as grids
import pygeogrids.netcdf as ncgrid
from pynetcf.time_series import GriddedNcContiguousRaggedTs

from ascat.read_native.cdr import AscatTimeSeries


class Ascat_data(object):

    """
    Class that provides access to ASCAT data stored in userformat which
    is available from the TU Wien FTP Server after registration at
    http://rs.geo.tuwien.ac.at.

    Parameters
    ----------
    path : string
        Path to data folder which contains the zip files from the FTP server
    grid_path : string
        Path to grid_info folder which contains txt files with information
        about grid point index,latitude, longitude and cell
    grid_info_filename : string, optional
        Name of the grid info txt file in grid_path
    advisory_flags_path : string, optional
        Path to advisory flags .dat files, if not provided they will
        not be used
    topo_threshold : int, optional
        If topographic complexity of read grid point is above this
        threshold a warning is output during reading
    wetland_threshold : int, optional
        If wetland fraction of read grid point is above this
        threshold a warning is output during reading

    Attributes
    ----------
    path : string
        Path to data folder which contains the zip files from the FTP server
    grid_path : string
        Path to grid_info folder which contains txt files with information
        about grid point index,latitude, longitude and cell
    grid_info_filename : string
        Name of the grid info txt file in grid_path
    grid_info_np_filename : string
        Name of the numpy save file to the grid information
    topo_threshold : int
        If topographic complexity of read grid point is above this
        threshold a warning is output during reading
    wetland_threshold : int
        If wetland fraction of read grid point is above this
        threshold a warning is output during reading
    grid_info_loaded : boolean
        True if the grid information has already been loaded
    grid : :class:`pygeogrids.grids.CellGrid` object
        CellGrid object, which provides nearest neighbor search and
        other features
    advisory_flags_path : string
        Path to advisory flags .dat files, if not provided they will
        not be used
    include_advflags : boolean
        True if advisory flags are available

    Methods
    -------
    unzip_cell(cell)
        unzips zipped grid point files into subdirectory
    read_advisory_flags(gpi)
        reads the advisory flags for a given grid point index
    """

    def __init__(self, path, grid_path,
                 grid_info_filename='TUW_W54_01_lonlat-ld-land.txt',
                 advisory_flags_path=None, topo_threshold=50,
                 wetland_threshold=50):
        """
        Set path and threshold
        """
        self.path = path
        self.grid_path = grid_path
        self.grid_info_filename = grid_info_filename
        self.grid_info_np_filename = 'TUW_W54_01_lonlat-ld-land.npy'
        self.topo_threshold = topo_threshold
        self.wetland_threshold = wetland_threshold
        self.grid_info_loaded = False
        self.grid = None
        self.advisory_flags_path = advisory_flags_path
        if self.advisory_flags_path is None:
            self.include_advflags = False
        else:
            self.include_advflags = True
            self.adv_flags_struct = np.dtype([('gpi', np.int32),
                                              ('snow', np.uint8, 366),
                                              ('frozen', np.uint8, 366),
                                              ('water', np.uint8),
                                              ('topo', np.uint8)])

    def _load_grid_info(self):
        """
        Reads the grid info for all land points from the txt file provided
        by TU Wien. The first time the actual txt file is parsed and saved
        as a numpy array to speed up future data access.
        """
        grid_info_np_filepath = os.path.join(
            self.grid_path, self.grid_info_np_filename)

        if os.path.exists(grid_info_np_filepath):
            grid_info = np.load(grid_info_np_filepath)

        else:
            grid_info_filepath = os.path.join(
                self.grid_path, self.grid_info_filename)
            grid_info = np.loadtxt(
                grid_info_filepath, delimiter=',', skiprows=1)
            np.save(os.path.join(self.grid_path,
                                 self.grid_info_np_filename), grid_info)

        self.grid = grids.CellGrid(grid_info[:, 2], grid_info[:, 1],
                                   grid_info[:, 3].astype(np.int16),
                                   gpis=grid_info[:, 0].astype(np.int32))

        self.grid_info_loaded = True

    def unzip_cell(self, cell):
        """
        Unzips the downloaded .zip cell file into the directory of
        os.path.join(self.path, cell).

        Parameters
        ----------
        cell : int
            cell number
        """
        filepath = os.path.join(self.path, '%4d.zip' % cell)
        unzip_file_path = os.path.join(self.path, '%4d' % cell)

        if not os.path.exists(unzip_file_path):
            os.mkdir(unzip_file_path)

        zfile = zipfile.ZipFile(filepath)
        for name in zfile.namelist():
            (dirname, filename) = os.path.split(name)
            fd = open(os.path.join(unzip_file_path, filename), "w")
            fd.write(zfile.read(name))
            fd.close()
        zfile.close()

    def _datetime_arr(self, longdate):
        """
        Parsing function that takes a number of type long which contains
        YYYYMMDDHH and returns a datetime object

        Parameters
        ----------
        longdate : long
            Date including hour as number of type long in format YYYYMMDDHH

        Returns
        -------
        datetime : datetime
        """
        string = str(longdate)
        year = int(string[0:4])
        month = int(string[4:6])
        day = int(string[6:8])
        hour = int(string[8:])

        return datetime(year, month, day, hour)

    def _read_ts(self, *args, **kwargs):
        """
        Takes either 1 or 2 arguments and calls the correct function
        which is either reading the gpi directly or finding
        the nearest gpi from given lat,lon coordinates and then reading it
        """

        if not self.grid_info_loaded:
            self._load_grid_info()

        if len(args) == 1:
            return self._read_gp(args[0], **kwargs)
        if len(args) == 2:
            return self._read_lonlat(args[0], args[1], **kwargs)

    def _read_gp(self, gpi, **kwargs):
        """
        Reads the time series of the given grid point index. Masks frozen
        and snow observations if keywords are present

        Parameters
        ----------
        gpi : long
            grid point index
        mask_frozen_prob : int,optional
            if included in kwargs then all observations taken when
            frozen probability > mask_frozen_prob are removed from the result
        mask_snow_prob : int,optional
            if included in kwargs then all observations taken when
            snow probability > mask_snow_prob are removed from the result

        Returns
        -------
        df : pandas.DataFrame
            containing all fields in the list self.include_in_df
            plus frozen_prob and snow_prob if a path to advisory flags was
            set during initialization
        """
        cell = self.grid.gpi2cell(int(gpi))

        gp_file = os.path.join(
            self.path, '%4d' % cell, self.gp_filename_template % gpi)

        if not os.path.exists(gp_file):
            print('first time reading from cell %4d unzipping ...' % cell)
            self.unzip_cell(cell)

        data = np.fromfile(gp_file, dtype=self.gp_filestruct)
        dates = data['DAT']

        datetime_parser = np.vectorize(self._datetime_arr)

        datetimes_correct = datetime_parser(dates)

        dict_df = {}

        for into_df in self.include_in_df:
            d = np.ma.asarray(data[into_df], dtype=self.datatype[into_df])
            d = np.ma.masked_equal(d, self.nan_values[into_df])
            if into_df in self.scale_factor.keys():
                d = d * self.scale_factor[into_df]
            dict_df[into_df] = d

        df = pd.DataFrame(dict_df, index=datetimes_correct)

        if self.include_advflags:
            adv_flags, topo, wetland = self.read_advisory_flags(gpi)

            if topo >= self.topo_threshold:
                warnings.warn(
                    "Warning gpi shows topographic complexity of %d %%. Data might not be usable." % topo)
            if wetland >= self.wetland_threshold:
                warnings.warn(
                    "Warning gpi shows wetland fraction of %d %%. Data might not be usable." % wetland)

            df['doy'] = df.index.dayofyear
            df = df.join(adv_flags, on='doy', how='left')
            del df['doy']

            if 'mask_frozen_prob' in kwargs:
                mask_frozen = kwargs['mask_frozen_prob']
                df = df[df['frozen_prob'] <= mask_frozen]

            if 'mask_snow_prob' in kwargs:
                mask_snow = kwargs['mask_snow_prob']
                df = df[df['snow_prob'] <= mask_snow]

        lon, lat = self.grid.gpi2lonlat(gpi)

        return df, gpi, lon, lat, cell

    def _read_lonlat(self, lon, lat, **kwargs):
        """
        Get nearest grid point and read data.

        Parameters
        ----------
        lon : float
          Longitude.
        lat : float
          Latitude.

        Returns
        -------
        data : numpy.ndarray
          Read data.
        """
        gpi, dist = self.grid.find_nearest_gpi(lon, lat)
        return self._read_gp(int(gpi), **kwargs)

    def read_advisory_flags(self, gpi):
        """
        Read the advisory flags located in the self.advisory_flags_path
        Advisory flags include frozen probability, snow cover probability
        topographic complexity and wetland fraction.

        Parameters
        ----------
        gpi : long
            grid point index

        Returns
        -------
        df : pandas.DataFrame
            containing the columns frozen_prob and snow_prob. lenght 366
            with one entry for every day of the year, including February 29th
        topo : numpy.uint8
            topographic complexity ranging from 0-100
        wetland : numpy.uint8
            wetland fraction of pixel in percent
        """
        if not self.include_advflags:
            raise RuntimeError("Error: advisory_flags_path is not set")

        if not self.grid_info_loaded:
            self._load_grid_info()

        cell = self.grid.gpi2cell(gpi)
        adv_file = os.path.join(
            self.advisory_flags_path, '%d_advisory-flags.dat' % cell)
        data = np.fromfile(adv_file, dtype=self.adv_flags_struct)
        index = np.where(data['gpi'] == gpi)[0]

        data = data[index]

        snow = data['snow'][0]
        snow[snow == 0] += 101
        snow -= 101

        df = pd.DataFrame({'snow_prob': snow,
                           'frozen_prob': data['frozen'][0]})

        return df, data['topo'][0], data['water'][0]


class Ascat_SSM(Ascat_data):

    """
    Class for reading ASCAT SSM data. It extends Ascat_data and provides the
    information necessary for reading SSM data

    Parameters
    ----------
    path : string
        Path to data folder which contains the zip files from the FTP server
    grid_path : string
        Path to grid_info folder which contains txt files with information
        about grid point index,latitude, longitude and cell
    grid_info_filename : string, optional
        Name of the grid info txt file in grid_path
    advisory_flags_path : string, optional
        Path to advisory flags .dat files, if not provided they will not
        be used
    topo_threshold : int, optional
        If topographic complexity of read grid point is above this
        threshold a warning is output during reading
    wetland_threshold : int, optional
        If wetland fraction of read grid point is above this
        threshold a warning is output during reading

    Attributes
    ----------
    gp_filename_template : string
        Defines how the gpi is put into the template string to make the
        filename
    gp_filestruct : numpy.dtype
        Structure template of the SSM .dat file
    scale_factor : dict
        Factor by which to multiply the raw data to get the correct values
        for each field in the gp_filestruct
    include_in_df : list
        List of fields that should be returned to the user after reading
    nan_values : dict
        Nan value saved in the file which will be replaced by numpy.nan values
        during reading
    datatype : dict
        Datatype of the fields that the return data should have

    Methods
    -------
    read_ssm(*args,**kwargs)
        read surface soil moisture
    """

    def __init__(self, *args, **kwargs):
        super(Ascat_SSM, self).__init__(*args, **kwargs)
        self.gp_filename_template = 'TUW_ASCAT_SSM_W55_gp%d.dat'
        self.gp_filestruct = np.dtype([('DAT', np.int32),
                                       ('SSM', np.uint8),
                                       ('ERR', np.uint8),
                                       ('SSF', np.uint8)])

        self.scale_factor = {'SSM': 0.5, 'ERR': 0.5}
        self.include_in_df = ['SSM', 'ERR', 'SSF']
        self.nan_values = {'SSM': 255, 'ERR': 255, 'SSF': 255}
        self.datatype = {'SSM': np.float, 'ERR': np.float, 'SSF': np.int}

    def read_ssm(self, *args, **kwargs):
        """
        Function to read SSM takes either 1 or 2 arguments.
        It can be called as read_ssm(gpi, **kwargs) or
        read_ssm(lon, lat, **kwargs)

        Parameters
        ----------
        gpi : int
            Grid point index
        lon : float
            Longitude of point
        lat : float
            Latitude of point
        mask_ssf : boolean, optional
            Default False, if True only SSF values of 1 will be allowed,
            all others are removed
        mask_frozen_prob : int,optional
            If included in kwargs then all observations taken when
            frozen probability > mask_frozen_prob are removed from the result
        mask_snow_prob : int,optional
            If included in kwargs then all observations taken when
            snow probability > mask_snow_prob are removed from the result

        Returns
        -------
        AscatTimeSeries : object
            :class:`ascat.AscatTimeSeries` instance
        """
        df, gpi, lon, lat, cell = super(
            Ascat_SSM, self)._read_ts(*args, **kwargs)
        if 'mask_ssf' in kwargs:
            mask_ssf = kwargs['mask_ssf']
            if mask_ssf:
                df = df[df['SSF'] == 1]

        return AscatTimeSeries(gpi, lon, lat, cell, df)


class Ascat_SWI(Ascat_data):

    """
    Class for reading ASCAT SWI data. It extends Ascat_data and provides the
    information necessary for reading SWI data

    Parameters
    ----------
    path : string
        Path to data folder which contains the zip files from the FTP server
    grid_path : string
        Path to grid_info folder which contains txt files with information
        about grid point index,latitude, longitude and cell
    grid_info_filename : string, optional
        Name of the grid info txt file in grid_path
    advisory_flags_path : string, optional
        Path to advisory flags .dat files, if not provided they will not
        be used
    topo_threshold : int, optional
        If topographic complexity of read grid point is above this
        threshold a warning is output during reading
    wetland_threshold : int, optional
        If wetland fraction of read grid point is above this
        threshold a warning is output during reading

    Attributes
    ----------
    gp_filename_template : string
        Defines how the gpi is put into the template string to make
        the filename
    gp_filestruct : numpy.dtype
        Structure template of the SSM .dat file
    scale_factor : dict
        Factor by which to multiply the raw data to get the correct values
        for each field in the gp_filestruct
    include_in_df : list
        List of fields that should be returned to the user after reading
    nan_values : dict
        Nan value saved in the file which will be replaced by numpy.nan values
        during reading
    datatype : dict
        Datatype of the fields that the return data should have
    T_SWI : dict
        Information about which numerical T-Value maps to which entry in the
        datastructure
    T_QFLAG : dict
        Information about which numerical T-Value maps to which entry in the
        datastructure

    Methods
    -------
    read_swi(*args,**kwargs)
        read soil water index
    """

    def __init__(self, *args, **kwargs):
        super(Ascat_SWI, self).__init__(*args, **kwargs)
        self.gp_filename_template = 'TUW_ASCAT_SWI_W55_gp%d.dat'
        self.gp_filestruct = np.dtype([('DAT', np.int32),
                                       ('SWI_T=1', np.uint8),
                                       ('SWI_T=5', np.uint8),
                                       ('SWI_T=10', np.uint8),
                                       ('SWI_T=15', np.uint8),
                                       ('SWI_T=20', np.uint8),
                                       ('SWI_T=40', np.uint8),
                                       ('SWI_T=60', np.uint8),
                                       ('SWI_T=100', np.uint8),
                                       ('QFLAG_T=1', np.uint8),
                                       ('QFLAG_T=5', np.uint8),
                                       ('QFLAG_T=10', np.uint8),
                                       ('QFLAG_T=15', np.uint8),
                                       ('QFLAG_T=20', np.uint8),
                                       ('QFLAG_T=40', np.uint8),
                                       ('QFLAG_T=60', np.uint8),
                                       ('QFLAG_T=100', np.uint8)])

        self.scale_factor = {'SWI_T=1': 0.5, 'SWI_T=5': 0.5,
                             'SWI_T=10': 0.5, 'SWI_T=15': 0.5,
                             'SWI_T=20': 0.5, 'SWI_T=40': 0.5,
                             'SWI_T=60': 0.5, 'SWI_T=100': 0.5,
                             'QFLAG_T=1': 0.5, 'QFLAG_T=5': 0.5,
                             'QFLAG_T=10': 0.5, 'QFLAG_T=15': 0.5,
                             'QFLAG_T=20': 0.5, 'QFLAG_T=40': 0.5,
                             'QFLAG_T=60': 0.5, 'QFLAG_T=100': 0.5}

        self.include_in_df = ['SWI_T=1', 'SWI_T=5', 'SWI_T=10', 'SWI_T=15',
                              'SWI_T=20', 'SWI_T=40', 'SWI_T=60', 'SWI_T=100',
                              'QFLAG_T=1', 'QFLAG_T=5', 'QFLAG_T=10',
                              'QFLAG_T=15', 'QFLAG_T=20', 'QFLAG_T=40',
                              'QFLAG_T=60', 'QFLAG_T=100']

        self.nan_values = {'SWI_T=1': 255, 'SWI_T=5': 255, 'SWI_T=10': 255,
                           'SWI_T=15': 255, 'SWI_T=20': 255, 'SWI_T=40': 255,
                           'SWI_T=60': 255, 'SWI_T=100': 255,
                           'QFLAG_T=1': 255, 'QFLAG_T=5': 255,
                           'QFLAG_T=10': 255, 'QFLAG_T=15': 255,
                           'QFLAG_T=20': 255, 'QFLAG_T=40': 255,
                           'QFLAG_T=60': 255, 'QFLAG_T=100': 255}

        self.datatype = {'SWI_T=1': np.float, 'SWI_T=5': np.float,
                         'SWI_T=10': np.float, 'SWI_T=15': np.float,
                         'SWI_T=20': np.float, 'SWI_T=40': np.float,
                         'SWI_T=60': np.float, 'SWI_T=100': np.float,
                         'QFLAG_T=1': np.float, 'QFLAG_T=5': np.float,
                         'QFLAG_T=10': np.float, 'QFLAG_T=15': np.float,
                         'QFLAG_T=20': np.float, 'QFLAG_T=40': np.float,
                         'QFLAG_T=60': np.float, 'QFLAG_T=100': np.float}

        self.T_SWI = {1: 'SWI_T=1', 5: 'SWI_T=5', 10: 'SWI_T=10',
                      15: 'SWI_T=15', 20: 'SWI_T=20', 40: 'SWI_T=40',
                      60: 'SWI_T=60', 100: 'SWI_T=100'}

        self.T_QFLAG = {1: 'QFLAG_T=1', 5: 'QFLAG_T=5', 10: 'QFLAG_T=10',
                        15: 'QFLAG_T=15', 20: 'QFLAG_T=20', 40: 'QFLAG_T=40',
                        60: 'QFLAG_T=60', 100: 'QFLAG_T=100'}

    def read_swi(self, *args, **kwargs):
        """
        Function to read SWI takes either 1 or 2 arguments being.
        It can be called as read_swi(gpi, **kwargs) or
        read_swi(lon, lat, **kwargs)

        Parameters
        ----------
        gpi : int
            grid point index
        lon : float
            longitude of point
        lat : float
            latitude of point
        T : int, optional
            if set only the SWI and QFLAG of this T-Value will be returned
        mask_qf : int, optional
            if set, SWI values with a QFLAG value lower than the mask_qf
            value will be masked. This is done for each T value independently
        mask_frozen_prob : int,optional
            if included in kwargs then all observations taken when
            frozen probability > mask_frozen_prob are removed from the result
        mask_snow_prob : int,optional
            if included in kwargs then all observations taken when
            snow probability > mask_snow_prob are removed from the result
        Returns
        -------
        df : pandas.DataFrame
            containing all fields in self.include_in_df plus frozen_prob
            and snow_prob if advisory_flags_path was set. If T was set then
            only SWI and QFLAG values for the selected T value are included
            plut frozen_prob and snow_prob if applicable
        """
        df, gpi, lon, lat, cell = super(Ascat_SWI, self)._read_ts(*args,
                                                                  **kwargs)

        if 'T' in kwargs:
            T = kwargs['T']
            if T in self.T_SWI.keys():
                if self.include_advflags:
                    df = df[[self.T_SWI[T], self.T_QFLAG[T],
                             'frozen_prob', 'snow_prob']]
                else:
                    df = df[[self.T_SWI[T], self.T_QFLAG[T]]]
            else:
                raise RuntimeError(
                    "Invalid T value. Choose one of " + str(sorted(self.T_SWI.keys())))

            # remove rows that have to small QFLAG
            if 'mask_qf' in kwargs:
                mask_qf = kwargs['mask_qf']
                if mask_qf:
                    df = df[df[self.T_QFLAG[T]] >= mask_qf]

        else:
            # mask each T value according to qf threshold
            if 'mask_qf' in kwargs:
                mask_qf = kwargs['mask_qf']
                for key in self.T_SWI:
                    masked = df[self.T_QFLAG[key]] <= mask_qf
                    df[self.T_SWI[key]][masked] = np.NAN
                    df[self.T_QFLAG[key]][masked] = np.NAN

        return AscatTimeSeries(gpi, lon, lat, cell, df)


class AscatVodTs(GriddedNcContiguousRaggedTs):

    """
    Class that provides access to ASCAT VOD data stored in netCDF format.

    Parameters
    ----------
    path : string
        Path to data folder which contains the zip files from the FTP server
    grid_path : string
        Path to grid_info folder which contains a netcdf file with
        information about grid point index,latitude, longitude and cell
    grid_info_filename : string, optional
        Name of the grid info netCDF file in grid_path
        default 'TUW_WARP5_grid_info_2_1.nc'

    Attributes
    ----------
    path : string
        Path to data folder which contains the zip files from the FTP server
    grid_path : string
        Path to grid_info folder which contains txt files with information
        about grid point index,latitude, longitude and cell
    grid_info_filename : string, optional
        Name of the grid info netCDF file in grid_path
        default 'TUW_WARP5_grid_info_2_1.nc'
    grid : grids.CellGrid object
        CellGrid object, which provides nearest neighbor search and
        other features
    variables : list of string
    """

    def __init__(self, path, grid_path,
                 grid_info_filename='TUW_WARP5_grid_info_2_1.nc',
                 variables=None):

        grid = ncgrid.load_grid(os.path.join(grid_path, grid_info_filename),
                                subset_flag='land')

        self.path = path
        self.grid_path = grid_path
        self.grid_info_filename = grid_info_filename

        self.variables = variables

        if self.variables is None:
            self.variables = ['vod']

        super(AscatVodTs, self).__init__(path, grid)
